# Copyright (c) 2011-2012 Cloudera, Inc. All rights reserved.

import re

from cmf.monitor.json_metrics_utils import ParseError, DictWithAttr

class JsonParserJmx(object):
  def __init__(self, json_dict, key, context={}):
    """
    May raise ParseError if the json structure is not what we expect. See the 
    _get_beans function for an explanation of how the context argument is used.
    The key argument should correspond to an entry in the BEANS dictionary in
    this file.
    """
    beans = BEANS[key]
    json_beans = _get_beans(json_dict, beans, context)
    for bean in beans:
      setattr(self, bean.get_alias(), json_beans.get(bean.get_alias()))

def _get_beans(json_dictionary, beans, context):
  """
  Given a Hadoop or Cloudera Manager JMX json dict, returns a map of beans found
  matching the input set of bean objects as a dictionary from bean alias to the
  bean as a dictionary of metric values. May raise ParseError.

  Returns a DictWithAttr.

  The beans argument is a collection of Bean objects that must implement two
  methods: matches(name, context) and get_alias().

  The context argument is an arbitrary dictionary passed through to the matches
  method invocation on the beans. It is used to pass configuration information
  from the monitor for which we are collecting into the parsing process. For
  example, we use it to pass port information for a NameNode to distinguish the
  RPC statistics for the client facing server vs. the service facing server.
  For example, consider the following:

     {
    "name" : "Hadoop:service=NameNode,name=RpcDetailedActivityForPort8020",
    "modelerType" : "RpcDetailedActivityForPort8020",
    ...
    "NumOpenConnections" : 5,
    ...
    }, {
    "name" : "Hadoop:service=NameNode,name=RpcDetailedActivityForPort8022",
    "modelerType" : "RpcDetailedActivityForPort8022",
    ...
    "NumOpenConnections" : 5,
    ...
    },

  If we want to distinguish these two beans with different metrics we need some
  way to know with which metrics we ought to associate them. To do this we use
  the PortAwareMetricsBean, leveraging the input context below.
  """
  bean_dict_list = json_dictionary.get('beans')
  if bean_dict_list is None:
    err = "Cannot find 'beans' list in input json"
    raise ParseError(err)

  if not isinstance(bean_dict_list, list):
    err = "Expecting a 'beans' list. Got %r" % (bean_dict_list, )
    raise ParseError(err)

  ret = {}
  for d in map(lambda x: DictWithAttr(x), bean_dict_list):
    name = d.get('name')
    for bean in beans:
      if bean.matches(name, context):
        ret[bean.get_alias()] = d
        break
    if len(ret.keys()) == len(beans):
      break

  return ret

class Bean:
  """
  Holds metadata used to represent a bean from which we collect metrics.
  """
  def __init__(self, regex, alias):
    # A regular expression used to identify the bean.
    self._regex = regex
    # The name we'll use internally for the bean.
    self._alias = alias

  # Whether the input bean name matches this bean defition. See _get_beans for
  # an explanation of the context parameter (which this class ignores entirely)
  def matches(self, name, context):
    return self._regex.match(name)

  # Whether the input bean name matches this bean definition. This will only be
  # called by metric-tool and so must work without context information.
  def matches_for_metric_tool(self, name):
    return self._regex.match(name)

  def get_alias(self):
    return self._alias

# This can be used to make a PortAwareMetricsBean match any port.
MATCH_ANY_PORT = "match-any-port"

class PortAwareMetricsBean:
  """
  Holds metadata used to represent a bean from which we collect metrics. This is
  distinct from Bean in that it can take a port number from a metrics access
  context to customize its extraction behavior. This bean assumes that the input
  regex has a single extraction pattern.
  """
  def __init__(self, regex, alias, port_name_in_context):
    # A regular expression used to identify the bean.
    self._regex = regex
    # The name we'll use internally for the bean.
    self._alias = alias
    # The name we use to lookup the port value in the access context.
    self._port_name_in_context = port_name_in_context

  # See Bean above. This method looks up the port we are supposed to match from
  # the input context via the name used in construction. That value is compared
  # to the value extracted from the input bean name. If there is a match, or if
  # context told us to MATCH_ANY_PORT we match, otherwise we do not. See the
  # example in _get_beans for more detail.
  def matches(self, name, context):
    port = context.get(self._port_name_in_context)
    if port is None:
      return False
    m = self._regex.match(name)
    if m is None:
      return False
    if port == MATCH_ANY_PORT:
      return True
    return port == int(m.group(1))

  def matches_for_metric_tool(self, name):
    return self._regex.match(name)

  def get_alias(self):
    return self._alias

_JAVA_MEMORY_BEAN = Bean(re.compile('java.lang:type=Memory$'), 'java_memory')
_EVENT_STATS_BEAN = \
    Bean(re.compile('com.cloudera.cmf.event.publish:type=EventStorePublisherWithRetryMXBean'),
         'event_stats')

def _make_jvm_metrics_bean(name):
  return Bean(re.compile('Hadoop:service=%s,name=JvmMetrics' % name), 'jvm')

def _make_ugi_metrics_bean(name):
  return Bean(re.compile('Hadoop:service=%s,name=UgiMetrics' % name), 'ugi')

def _make_rpc_metrics_bean(name):
  return Bean(re.compile('Hadoop:service=%s,name=RpcActivityForPort\d+' % name),
              'rpc')

def _make_detailed_rpc_metrics_bean(name):
  return Bean(re.compile('Hadoop:service=%s,name=RpcDetailedActivityForPort\d+' % name),
              'detailed_rpc')

def _make_port_aware_rpc_metrics_bean(name, alias, port_name_in_context):
  return PortAwareMetricsBean(
    re.compile('Hadoop:service=%s,name=RpcActivityForPort(\d+)' % name),
    alias,
    port_name_in_context)

def _make_port_aware_detailed_rpc_metrics_bean(name, alias, port_name_in_context):
  return PortAwareMetricsBean(
    re.compile('Hadoop:service=%s,name=RpcDetailedActivityForPort(\d+)' % name),
    alias,
    port_name_in_context)

def _make_metrics_control_metrics_bean(name):
  return Bean(re.compile('Hadoop:service=%s,name=MetricsSystem,sub=Control' % name),
              'metrics_control')

def _make_metrics_stats_metrics_bean(name):
  return Bean(re.compile('Hadoop:service=%s,name=MetricsSystem,sub=Stats' % name),
              'metrics_stats')

# The NameNode can have two RPC servers. See OPSAPS-14550 for more information.
NAMENODE_CLIENT_RPC_PORT_NAME = 'client_rpc_port'
NAMENODE_SERVICE_RPC_PORT_NAME = 'service_rpc_port'

BEANS = dict(
    NAMENODE=[
        _JAVA_MEMORY_BEAN,
        _make_jvm_metrics_bean('NameNode'),
        _make_port_aware_rpc_metrics_bean('NameNode',
                                          'rpc',
                                          NAMENODE_CLIENT_RPC_PORT_NAME),
        _make_port_aware_detailed_rpc_metrics_bean('NameNode',
                                                   'detailed_rpc',
                                                   NAMENODE_CLIENT_RPC_PORT_NAME),
        _make_detailed_rpc_metrics_bean('NameNode'),
        _make_ugi_metrics_bean('NameNode'),
        _make_metrics_control_metrics_bean('NameNode'),
        _make_metrics_stats_metrics_bean('NameNode'),
        Bean(re.compile('Hadoop:service=NameNode,name=NameNodeInfo'),
             'namenode_info'),
        Bean(re.compile('Hadoop:service=NameNode,name=NameNodeActivity'),
             'namenode_activity'),
        Bean(re.compile('Hadoop:service=NameNode,name=FSNamesystem$'),
             'fsnamesystem'),
        Bean(re.compile('Hadoop:service=NameNode,name=FSNamesystemState'), 
             'fsnamesystem_state'),
        Bean(re.compile('Hadoop:service=NameNode,name=IPCLoggerChannel'), 
             'ipc_logger_channel'),
        ],
    SECONDARYNAMENODE=[
        _JAVA_MEMORY_BEAN,
        _make_jvm_metrics_bean('SecondaryNameNode'),
        _make_rpc_metrics_bean('SecondaryNameNode'),
        _make_detailed_rpc_metrics_bean('SecondaryNameNode'),
        _make_ugi_metrics_bean('SecondaryNameNode'),
        _make_metrics_control_metrics_bean('SecondaryNameNode'),
        _make_metrics_stats_metrics_bean('SecondaryNameNode')
        ],
    DATANODE=[
        _JAVA_MEMORY_BEAN,
        _make_jvm_metrics_bean('DataNode'),
        _make_rpc_metrics_bean('DataNode'),
        _make_detailed_rpc_metrics_bean('DataNode'),
        _make_ugi_metrics_bean('DataNode'),
        _make_metrics_control_metrics_bean('DataNode'),
        _make_metrics_stats_metrics_bean('DataNode'),
        Bean(re.compile('Hadoop:service=DataNode,name=DataNodeActivity'), 
             'datanode_activity'),
        Bean(re.compile('Hadoop:service=DataNode,name=DataNodeInfo'), 
             'datanode_info'),
        Bean(re.compile('Hadoop:service=DataNode,name=FSDatasetState'), 
             'fs_dataset_state'),
        ],
    JOURNALNODE=[
        _JAVA_MEMORY_BEAN,
        _make_jvm_metrics_bean('JournalNode'),
        _make_rpc_metrics_bean('JournalNode'),
        _make_detailed_rpc_metrics_bean('JournalNode'),
        _make_ugi_metrics_bean('JournalNode'),
        _make_metrics_control_metrics_bean('JournalNode'),
        _make_metrics_stats_metrics_bean('JournalNode'),
        Bean(re.compile('Hadoop:service=JournalNode,name=Journal-.+'), 
             'journal')
        ],
    SERVICEMONITOR=[
        _EVENT_STATS_BEAN,
        Bean(re.compile("com.cloudera.cmf.event.publish:type=PipelineStageMXBean,stage=roleStageTSStore"), 
             'role_stage')             
        ],
    HOSTMONITOR=[
        _EVENT_STATS_BEAN,
        Bean(re.compile("com.cloudera.cmf.event.publish:type=PipelineStageMXBean,stage=hostStageTSStore"),
             'host_stage')
        ],
    ACTIVITYMONITOR=[
        Bean(re.compile("com.cloudera.cmf.event.publish:type=PipelineStageMXBean,stage=activityTree"), 
             'activity_tree_pipeline'),
        Bean(re.compile("com.cloudera.cmf.event.publish:type=PipelineStageMXBean,stage=activityMonitoring"), 
             'activity_monitor_pipeline'),
        Bean(re.compile("com.cloudera.enterprise:type=PeriodicEnterpriseServiceMXBean,service=ActivityAndAttemptPurgeService"), 
             'activity_monitor_purger')
        ],
    EVENTSERVER=[
        Bean(re.compile('com.cloudera.cmf.eventcatcher.server:type=EventCatcherServiceMXBean'), 
             'event_server'),
        Bean(re.compile("com.cloudera.cmf.event.publish:type=PipelineStageMXBean,stage=tagger-writer"), 
             'event_pipeline')
        ],
    NAVIGATOR=[
        Bean(re.compile("com.cloudera.navigator.ipc:type=NavigatorIPC,name=read-request"),
             'read_requests'),
        Bean(re.compile("com.cloudera.navigator.ipc:type=NavigatorIPC,name=insert-request"),
             'insert_requests'),
        Bean(re.compile("com.cloudera.navigator.ipc:type=NavigatorIPC,name=insertion-rate"),
             'insertion_rate'),
        Bean(re.compile("com.cloudera.navigator.ipc:type=NavigatorIPC,name=active-read-requests"),
             'active_read_requests'),
        Bean(re.compile("com.cloudera.navigator.ipc:type=NavigatorIPC,name=active-write-requests"),
             'active_write_requests')
        ],
    )

